# Do not modify this file by hand.
#
# It is automatically generated by src/api_wrappers/generatePythonAPIWrappers.py.
# (Run make api_wrappers to update it.)

from __future__ import print_function, unicode_literals, division, absolute_import

import sys

from dxpy import DXHTTPRequest
from dxpy.utils import Nonce

def analysis_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_terminate(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/terminate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-terminate
    """
    return DXHTTPRequest('/%s/terminate' % object_id, input_params, always_retry=always_retry, **kwargs)

def app_add_authorized_users(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-addauthorizedusers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_add_categories(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-addcategories
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_add_developers(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-adddevelopers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_add_tags(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-addtags
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_delete(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/delete API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-delete
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/delete' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_describe(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-describe
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/describe' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_get(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/get API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-get
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/get' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_install(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/install API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-install
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/install' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_list_authorized_users(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/listAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-listauthorizedusers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_list_categories(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/listCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-listcategories
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_list_developers(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/listDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-listdevelopers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_publish(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/publish API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-publish
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/publish' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_authorized_users(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removeauthorizedusers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_categories(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removecategories
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_developers(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removedevelopers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_tags(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removetags
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_run(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/run' % fully_qualified_version, input_params_cp, always_retry=always_retry, **kwargs)

def app_validate_batch(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/validateBatch API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-validatebatch
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/validateBatch' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_uninstall(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/uninstall API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-uninstall
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/uninstall' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_update(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-update
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/update' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/app/new', input_params_cp, always_retry=always_retry, **kwargs)

def applet_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_get(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/get API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-get
    """
    return DXHTTPRequest('/%s/get' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_validate_batch(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/validateBatch API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-validatebatch
    """
    return DXHTTPRequest('/%s/validateBatch' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_run(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/%s/run' % object_id, input_params_cp, always_retry=always_retry, **kwargs)

def applet_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/applet/new', input_params_cp, always_retry=always_retry, **kwargs)

def container_clone(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/clone API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-clone
    """
    return DXHTTPRequest('/%s/clone' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/containers-for-execution#api-method-container-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_destroy(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/destroy API method.
    """
    return DXHTTPRequest('/%s/destroy' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_list_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/listFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-listfolder
    """
    return DXHTTPRequest('/%s/listFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_move(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/move API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-move
    """
    return DXHTTPRequest('/%s/move' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_new_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/newFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-newfolder
    """
    return DXHTTPRequest('/%s/newFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_remove_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/removeFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removefolder
    """
    return DXHTTPRequest('/%s/removeFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_remove_objects(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/removeObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removeobjects
    """
    return DXHTTPRequest('/%s/removeObjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_rename_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/renameFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-renamefolder
    """
    return DXHTTPRequest('/%s/renameFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle#api-method-class-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/databases#api-method-database-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_relocate(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /database-xxxx/relocate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/databases#api-method-database-xxxx-relocate
    """
    return DXHTTPRequest('/%s/relocate' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_download_file(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/downloadFile API method.
    """
    return DXHTTPRequest('/%s/downloadFile' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_list_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/listFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-listfolder
    """
    return DXHTTPRequest('/%s/listFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_new(input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /dbcluster/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-new
    """
    return DXHTTPRequest('/dbcluster/new', input_params, always_retry=always_retry, **kwargs)

def dbcluster_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_start(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/start API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-start
    """
    return DXHTTPRequest('/%s/start' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_stop(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/stop API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-stop
    """
    return DXHTTPRequest('/%s/stop' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_terminate(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/terminate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-terminate
    """
    return DXHTTPRequest('/%s/terminate' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_download(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/download API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-download
    """
    return DXHTTPRequest('/%s/download' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_upload(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/upload API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-upload
    """
    return DXHTTPRequest('/%s/upload' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/file/new', input_params_cp, always_retry=always_retry, **kwargs)

def global_workflow_add_authorized_users(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-addauthorizedusers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_add_categories(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-addcategories
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_add_developers(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-adddevelopers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_add_tags(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-addtags
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_delete(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/delete API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-delete
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/delete' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_describe(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-describe
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/describe' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_list_authorized_users(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/listAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-listauthorizedusers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_list_categories(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/listCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-listcategories
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_list_developers(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/listDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-listdevelopers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_publish(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/publish API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-publish
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/publish' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_authorized_users(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removeauthorizedusers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_categories(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removecategories
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_developers(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removedevelopers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_tags(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removetags
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_run(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/run' % fully_qualified_version, input_params_cp, always_retry=always_retry, **kwargs)

def global_workflow_update(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-update
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/update' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/globalworkflow/new', input_params_cp, always_retry=always_retry, **kwargs)

def job_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_get_log(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /job-xxxx/getLog API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-getlog
    """
    return DXHTTPRequest('/%s/getLog' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_terminate(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/terminate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-terminate
    """
    return DXHTTPRequest('/%s/terminate' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_get_identity_token(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/getIdentityToken API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-getIdentityToken
    """
    return DXHTTPRequest('/%s/getIdentityToken' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/job/new', input_params_cp, always_retry=always_retry, **kwargs)

def notifications_get(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /notifications/get API method.
    """
    return DXHTTPRequest('/notifications/get', input_params, always_retry=always_retry, **kwargs)

def notifications_mark_read(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /notifications/markRead API method.
    """
    return DXHTTPRequest('/notifications/markRead', input_params, always_retry=always_retry, **kwargs)

def org_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_find_members(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/findMembers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-findmembers
    """
    return DXHTTPRequest('/%s/findMembers' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_find_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/findProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-findprojects
    """
    return DXHTTPRequest('/%s/findProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_find_apps(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/findApps API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-findapps
    """
    return DXHTTPRequest('/%s/findApps' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_invite(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/invite API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-invite
    """
    return DXHTTPRequest('/%s/invite' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_remove_member(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/removeMember API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-removemember
    """
    return DXHTTPRequest('/%s/removeMember' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_set_member_access(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/setMemberAccess API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-setmemberaccess
    """
    return DXHTTPRequest('/%s/setMemberAccess' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/org/new', input_params_cp, always_retry=always_retry, **kwargs)

def project_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_archive(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/archive API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-archive
    """
    return DXHTTPRequest('/%s/archive' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_unarchive(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/unarchive API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-unarchive
    """
    return DXHTTPRequest('/%s/unarchive' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_clone(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/clone API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-clone
    """
    return DXHTTPRequest('/%s/clone' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_decrease_permissions(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/decreasePermissions API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-decreasepermissions
    """
    return DXHTTPRequest('/%s/decreasePermissions' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_destroy(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/destroy API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-destroy
    """
    return DXHTTPRequest('/%s/destroy' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_invite(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/invite API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-invite
    """
    return DXHTTPRequest('/%s/invite' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_leave(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/leave API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-leave
    """
    return DXHTTPRequest('/%s/leave' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_list_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/listFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-listfolder
    """
    return DXHTTPRequest('/%s/listFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_move(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/move API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-move
    """
    return DXHTTPRequest('/%s/move' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_new_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/newFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-newfolder
    """
    return DXHTTPRequest('/%s/newFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_remove_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/removeFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removefolder
    """
    return DXHTTPRequest('/%s/removeFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_remove_objects(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/removeObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removeobjects
    """
    return DXHTTPRequest('/%s/removeObjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_rename_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/renameFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-renamefolder
    """
    return DXHTTPRequest('/%s/renameFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_transfer(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/transfer API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-transfer
    """
    return DXHTTPRequest('/%s/transfer' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_update_sponsorship(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/updateSponsorship API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-updatesponsorship
    """
    return DXHTTPRequest('/%s/updateSponsorship' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_new(input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-new
    """
    return DXHTTPRequest('/project/new', input_params, always_retry=always_retry, **kwargs)

def record_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle#api-method-class-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/records#api-method-record-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/records#api-method-record-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/record/new', input_params_cp, always_retry=always_retry, **kwargs)

def system_describe_data_objects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/describeDataObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-describedataobjects
    """
    return DXHTTPRequest('/system/describeDataObjects', input_params, always_retry=always_retry, **kwargs)

def system_describe_executions(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/describeExecutions API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-describeexecutions
    """
    return DXHTTPRequest('/system/describeExecutions', input_params, always_retry=always_retry, **kwargs)

def system_describe_projects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/describeProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-describeprojects
    """
    return DXHTTPRequest('/system/describeProjects', input_params, always_retry=always_retry, **kwargs)

def system_find_affiliates(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findAffiliates API method.
    """
    return DXHTTPRequest('/system/findAffiliates', input_params, always_retry=always_retry, **kwargs)

def system_find_apps(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findApps API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findapps
    """
    return DXHTTPRequest('/system/findApps', input_params, always_retry=always_retry, **kwargs)

def system_find_data_objects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findDataObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-finddataobjects
    """
    return DXHTTPRequest('/system/findDataObjects', input_params, always_retry=always_retry, **kwargs)

def system_find_global_workflows(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findGlobalWorkflows API method.
    """
    return DXHTTPRequest('/system/findGlobalWorkflows', input_params, always_retry=always_retry, **kwargs)

def system_resolve_data_objects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/resolveDataObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-resolvedataobjects
    """
    return DXHTTPRequest('/system/resolveDataObjects', input_params, always_retry=always_retry, **kwargs)

def system_find_executions(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findExecutions API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findexecutions
    """
    return DXHTTPRequest('/system/findExecutions', input_params, always_retry=always_retry, **kwargs)

def system_find_analyses(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findAnalyses API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findanalyses
    """
    return DXHTTPRequest('/system/findAnalyses', input_params, always_retry=always_retry, **kwargs)

def system_find_databases(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findDatabases API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-finddatabases
    """
    return DXHTTPRequest('/system/findDatabases', input_params, always_retry=always_retry, **kwargs)

def system_find_jobs(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findJobs API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findjobs
    """
    return DXHTTPRequest('/system/findJobs', input_params, always_retry=always_retry, **kwargs)

def system_find_projects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findprojects
    """
    return DXHTTPRequest('/system/findProjects', input_params, always_retry=always_retry, **kwargs)

def system_find_users(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findUsers API method.
    """
    return DXHTTPRequest('/system/findUsers', input_params, always_retry=always_retry, **kwargs)

def system_find_project_members(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findProjectMembers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findprojectmembers
    """
    return DXHTTPRequest('/system/findProjectMembers', input_params, always_retry=always_retry, **kwargs)

def system_find_orgs(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findOrgs API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findorgs
    """
    return DXHTTPRequest('/system/findOrgs', input_params, always_retry=always_retry, **kwargs)

def system_generate_batch_inputs(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/generateBatchInputs API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-generatebatchinputs
    """
    return DXHTTPRequest('/system/generateBatchInputs', input_params, always_retry=always_retry, **kwargs)

def system_global_search(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/globalSearch API method.
    """
    return DXHTTPRequest('/system/globalSearch', input_params, always_retry=always_retry, **kwargs)

def system_greet(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/greet API method.
    """
    return DXHTTPRequest('/system/greet', input_params, always_retry=always_retry, **kwargs)

def system_headers(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/headers API method.
    """
    return DXHTTPRequest('/system/headers', input_params, always_retry=always_retry, **kwargs)

def system_shorten_url(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/shortenURL API method.
    """
    return DXHTTPRequest('/system/shortenURL', input_params, always_retry=always_retry, **kwargs)

def system_whoami(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/whoami API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-whoamiwiki.
    """
    return DXHTTPRequest('/system/whoami', input_params, always_retry=always_retry, **kwargs)

def user_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /user-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/users#api-method-user-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def user_update(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /user-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/users#api-method-user-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_add_stage(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/addStage API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-addstage
    """
    return DXHTTPRequest('/%s/addStage' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle#api-method-class-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_dry_run(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/dryRun API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-dryrun
    """
    return DXHTTPRequest('/%s/dryRun' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_is_stage_compatible(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/isStageCompatible API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-isstagecompatible
    """
    return DXHTTPRequest('/%s/isStageCompatible' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_move_stage(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/moveStage API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-movestage
    """
    return DXHTTPRequest('/%s/moveStage' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_overwrite(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/overwrite API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-overwrite
    """
    return DXHTTPRequest('/%s/overwrite' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_remove_stage(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/removeStage API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-removestage
    """
    return DXHTTPRequest('/%s/removeStage' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_run(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/%s/run' % object_id, input_params_cp, always_retry=always_retry, **kwargs)

def workflow_validate_batch(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/validateBatch API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-validatebatch
    """
    return DXHTTPRequest('/%s/validateBatch' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_update_stage_executable(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/updateStageExecutable API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-updatestageexecutable
    """
    return DXHTTPRequest('/%s/updateStageExecutable' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/workflow/new', input_params_cp, always_retry=always_retry, **kwargs)

